@@ -29,6 +29,9 @@ import org.aspectj.org.eclipse.jdt.core.Signature;
 import org.aspectj.org.eclipse.jdt.core.compiler.CharOperation;
 import org.aspectj.org.eclipse.jdt.internal.compiler.env.IBinaryMethod;
 import org.aspectj.org.eclipse.jdt.internal.compiler.env.IBinaryType;
+import org.aspectj.org.eclipse.jdt.internal.compiler.lookup.Binding;
+import org.aspectj.org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
+import org.aspectj.org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
 import org.aspectj.org.eclipse.jdt.internal.core.BinaryType;
 import org.aspectj.org.eclipse.jdt.internal.core.JavaElement;
 import org.aspectj.org.eclipse.jdt.internal.core.JavaModelManager;
@@ -133,6 +136,13 @@ public class InternalCompletionProposal extends CompletionProposal {
 	 */
 	private char[] name = null;
 
+	/**
+	 * Binding of the method or constructor being proposed, or
+	 * <code>null</code> if none.
+	 * Defaults to null.
+	 */
+	private Binding binding = null;
+
 	/**
 	 * Signature of the method, field type, member type,
 	 * relevant in the context, or <code>null</code> if none.
@@ -906,6 +916,45 @@ public class InternalCompletionProposal extends CompletionProposal {
 		this.name = name;
 	}
 
+	/**
+	 * Returns a binding of the method or field corresponding to this proposal or <code>null</code> if none.
+	 * <p>
+	 * The binding <em>may</em> be available for the following kinds of completion proposals:
+	 * <ul>
+	 *  <li><code>ANONYMOUS_CLASS_CONSTRUCTOR_INVOCATION</code> - {@link MethodBinding}
+	 * of the constructor being proposed</li>
+	 *  <li><code>CONSTRUCTOR_INVOCATION</code> - {@link MethodBinding}
+	 * of the constructor being proposed</li>
+	 *  <li><code>FIELD_REF</code> - {@link FieldBinding}
+	 * of the field being proposed</li>
+	 *  <li><code>FIELD_REF_WITH_CASTED_RECEIVER</code> - {@link FieldBinding}
+	 * of the field being proposed</li>
+	 *  <li><code>JAVADOC_FIELD_REF</code> - {@link FieldBinding}
+	 * of the field being proposed</li>
+	 *  <li><code>JAVADOC_METHOD_REF</code> - {@link MethodBinding}
+	 * of the method or constructor being proposed</li>
+	 *  <li><code>METHOD_DECLARATION</code> - {@link MethodBinding}
+	 * of the method or constructor being proposed</li>
+	 *  <li><code>METHOD_NAME_REFERENCE</code> - {@link MethodBinding}
+	 * of the method or constructor being proposed</li>
+	 *  <li><code>METHOD_REF</code> - {@link MethodBinding}
+	 * of the method or constructor being proposed</li>
+	 *  <li><code>METHOD_REF_WITH_CASTED_RECEIVER</code> - {@link MethodBinding}
+	 * of the method or constructor being proposed</li>
+	 * </ul>
+	 * For other kinds of completion proposals, this method returns <code>null</code>.
+	 * </p>
+	 *
+	 * @return the binding corresponding to this proposal (if available), or <code>null</code> if none
+	 */
+	public Binding getBinding() {
+		return this.binding;
+	}
+
+	public void setBinding(Binding binding) {
+		this.binding = binding;
+	}
+
 	/**
 	 * Returns the signature of the method or type
 	 * relevant in the context, or <code>null</code> if none.
