@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2000, 2013 IBM Corporation and others.
+ * Copyright (c) 2000, 2015 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -15,6 +15,7 @@
  *								bug 368546 - [compiler][resource] Avoid remaining false positives found when compiling the Eclipse SDK
  *								bug 345305 - [compiler][null] Compiler misidentifies a case of "variable can only be null"
  *								Bug 414380 - [compiler][internal] QualifiedNameReference#indexOfFirstFieldBinding does not point to the first field
+ *								Bug 458396 - NPE in CodeStream.invoke()
  *     Jesper S Moller - Contributions for
  *								bug 382721 - [1.8][compiler] Effectively final variables needs special treatment
  *								bug 331649 - [compiler][null] consider null annotations for fields
@@ -726,10 +727,10 @@ public TypeBinding getOtherFieldBindings(BlockScope scope) {
 	TypeBinding type = ((VariableBinding) this.binding).type;
 	int index = this.indexOfFirstFieldBinding;
 	if (index == length) { //	restrictiveFlag == FIELD
-		this.constant = ((FieldBinding) this.binding).constant();
+		this.constant = ((FieldBinding) this.binding).constant(scope);
 		// perform capture conversion if read access
 		return (type != null && (this.bits & ASTNode.IsStrictlyAssigned) == 0)
-				? type.capture(scope, this.sourceEnd)
+				? type.capture(scope, this.sourceStart, this.sourceEnd)
 				: type;
 	}
 	// allocation of the fieldBindings array	and its respective constants
@@ -738,7 +739,7 @@ public TypeBinding getOtherFieldBindings(BlockScope scope) {
 	this.otherDepths = new int[otherBindingsLength];
 
 	// fill the first constant (the one of the binding)
-	this.constant = ((VariableBinding) this.binding).constant();
+	this.constant = ((VariableBinding) this.binding).constant(scope);
 	// save first depth, since will be updated by visibility checks of other bindings
 	int firstDepth = (this.bits & ASTNode.DepthMASK) >> ASTNode.DepthSHIFT;
 	// iteration on each field
@@ -749,7 +750,7 @@ public TypeBinding getOtherFieldBindings(BlockScope scope) {
 
 		this.bits &= ~ASTNode.DepthMASK; // flush previous depth if any
 		FieldBinding previousField = field;
-		field = scope.getField(type.capture(scope, (int)this.sourcePositions[index]), token, this);
+		field = scope.getField(type.capture(scope, (int) (this.sourcePositions[index] >>> 32), (int)this.sourcePositions[index]), token, this);
 		int place = index - this.indexOfFirstFieldBinding;
 		this.otherBindings[place] = field;
 		this.otherDepths[place] = (this.bits & ASTNode.DepthMASK) >> ASTNode.DepthSHIFT;
@@ -770,7 +771,7 @@ public TypeBinding getOtherFieldBindings(BlockScope scope) {
 			}
 			// constant propagation can only be performed as long as the previous one is a constant too.
 			if (this.constant != Constant.NotAConstant) {
-				this.constant = field.constant();
+				this.constant = field.constant(scope);
 			}
 
 			if (field.isStatic()) {
@@ -787,7 +788,7 @@ public TypeBinding getOtherFieldBindings(BlockScope scope) {
 					}					
 					// check if accessing enum static field in initializer
 					if ((TypeBinding.equalsEquals(sourceType, declaringClass) || TypeBinding.equalsEquals(sourceType.superclass, declaringClass)) // enum constant body
-							&& field.constant() == Constant.NotAConstant
+							&& field.constant(scope) == Constant.NotAConstant
 							&& !methodScope.isStatic
 							&& methodScope.isInsideInitializerOrConstructor()) {
 						scope.problemReporter().enumStaticFieldUsedDuringInitialization(field, this);
@@ -813,7 +814,7 @@ public TypeBinding getOtherFieldBindings(BlockScope scope) {
 	type = (this.otherBindings[otherBindingsLength - 1]).type;
 	// perform capture conversion if read access
 	return (type != null && (this.bits & ASTNode.IsStrictlyAssigned) == 0)
-			? type.capture(scope, this.sourceEnd)
+			? type.capture(scope, this.sourceStart, this.sourceEnd)
 			: type;
 }
 
@@ -912,7 +913,7 @@ public void manageSyntheticAccessIfNecessary(BlockScope currentScope, FieldBindi
 	
 	if ((flowInfo.tagBits & FlowInfo.UNREACHABLE_OR_DEAD) != 0) return;
 	// index == 0 denotes the first fieldBinding, index > 0 denotes one of the 'otherBindings', index < 0 denotes a write access (to last binding)
-	if (fieldBinding.constant() != Constant.NotAConstant)
+	if (fieldBinding.constant(currentScope) != Constant.NotAConstant)
 		return;
 
 	if (fieldBinding.isPrivate()) { // private access
@@ -1079,7 +1080,7 @@ public TypeBinding resolveType(BlockScope scope) {
 						// check if accessing enum static field in initializer
 						if (declaringClass.isEnum()) {
 							if ((TypeBinding.equalsEquals(sourceType, declaringClass) || TypeBinding.equalsEquals(sourceType.superclass, declaringClass)) // enum constant body
-									&& fieldBinding.constant() == Constant.NotAConstant
+									&& fieldBinding.constant(scope) == Constant.NotAConstant
 									&& !methodScope.isStatic
 									&& methodScope.isInsideInitializerOrConstructor()) {
 								scope.problemReporter().enumStaticFieldUsedDuringInitialization(fieldBinding, this);
