@@ -29,12 +29,17 @@
  *								Bug 418537 - [1.8][null] Fix null type annotation analysis for poly conditional expressions
  *								Bug 428352 - [1.8][compiler] Resolution errors don't always surface
  *								Bug 429430 - [1.8] Lambdas and method reference infer wrong exception type with generics (RuntimeException instead of IOException)
+ *								Bug 435805 - [1.8][compiler][null] Java 8 compiler does not recognize declaration style null annotations
+ *								Bug 453483 - [compiler][null][loop] Improve null analysis for loops
+ *								Bug 455723 - Nonnull argument not correctly inferred in loop
  *        Andy Clement - Contributions for
  *                          Bug 383624 - [1.8][compiler] Revive code generation support for type annotations (from Olivier's work)
  *                          Bug 409250 - [1.8][compiler] Various loose ends in 308 code generation
  *******************************************************************************/
 package org.aspectj.org.eclipse.jdt.internal.compiler.ast;
 
+import org.aspectj.org.eclipse.jdt.core.compiler.CharOperation;
+import org.aspectj.org.eclipse.jdt.internal.compiler.ASTVisitor;
 import org.aspectj.org.eclipse.jdt.internal.compiler.classfmt.ClassFileConstants;
 import org.aspectj.org.eclipse.jdt.internal.compiler.codegen.*;
 import org.aspectj.org.eclipse.jdt.internal.compiler.flow.*;
@@ -77,6 +82,24 @@ public abstract class Statement extends ASTNode {
 	}
 public abstract FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo);
 
+/** Lambda shape analysis: *Assuming* this is reachable, analyze if this completes normally i.e control flow can reach the textually next statement.
+   For blocks, we don't perform intra-reachability analysis. We assume the lambda body is free of intrinsic control flow errors (if such errors
+   exist they will not be flagged by this analysis, but are guaranteed to surface later on.) 
+   
+   @see Block#doesNotCompleteNormally
+*/
+public boolean doesNotCompleteNormally() {
+	return false;
+}
+
+/** Lambda shape analysis: *Assuming* this is reachable, analyze if this completes by continuing i.e control flow cannot reach the textually next statement.
+    This is necessitated by the fact that continue claims to not complete normally. So this is necessary to discriminate between do { continue; } while (false); 
+    which completes normally and do { throw new Exception(); } while (false); which does not complete normally.
+*/
+public boolean completesByContinue() {
+	return false;
+}
+
 	public static final int NOT_COMPLAINED = 0;
 	public static final int COMPLAINED_FAKE_REACHABLE = 1;
 	public static final int COMPLAINED_UNREACHABLE = 2;
@@ -90,8 +113,7 @@ protected void analyseArguments(BlockScope currentScope, FlowContext flowContext
 		CompilerOptions compilerOptions = currentScope.compilerOptions();
 		if (compilerOptions.sourceLevel >= ClassFileConstants.JDK1_7 && methodBinding.isPolymorphic())
 			return;
-		boolean considerTypeAnnotations = compilerOptions.sourceLevel >= ClassFileConstants.JDK1_8
-				&& compilerOptions.isAnnotationBasedNullAnalysisEnabled;
+		boolean considerTypeAnnotations = currentScope.environment().usesNullTypeAnnotations();
 		boolean hasJDK15NullAnnotations = methodBinding.parameterNonNullness != null;
 		int numParamsToCheck = methodBinding.parameters.length;
 		int varArgPos = -1;
@@ -136,7 +158,7 @@ protected void analyseArguments(BlockScope currentScope, FlowContext flowContext
 					Expression argument = arguments[i];
 					int nullStatus = argument.nullStatus(flowInfo, flowContext); // slight loss of precision: should also use the null info from the receiver.
 					if (nullStatus != FlowInfo.NON_NULL) // if required non-null is not provided
-						flowContext.recordNullityMismatch(currentScope, argument, argument.resolvedType, expectedType, nullStatus);
+						flowContext.recordNullityMismatch(currentScope, argument, argument.resolvedType, expectedType, flowInfo, nullStatus, null);
 				}
 			}
 		} 
@@ -147,16 +169,16 @@ void analyseOneArgument18(BlockScope currentScope, FlowContext flowContext, Flow
 	if (argument instanceof ConditionalExpression && argument.isPolyExpression()) {
 		// drill into both branches using existing nullStatus per branch:
 		ConditionalExpression ce = (ConditionalExpression) argument;
-		ce.internalAnalyseOneArgument18(currentScope, flowContext, expectedType, ce.valueIfTrue, ce.ifTrueNullStatus, expectedNonNullness, originalExpected);
-		ce.internalAnalyseOneArgument18(currentScope, flowContext, expectedType, ce.valueIfFalse, ce.ifFalseNullStatus, expectedNonNullness, originalExpected);
+		ce.internalAnalyseOneArgument18(currentScope, flowContext, expectedType, ce.valueIfTrue, flowInfo, ce.ifTrueNullStatus, expectedNonNullness, originalExpected);
+		ce.internalAnalyseOneArgument18(currentScope, flowContext, expectedType, ce.valueIfFalse, flowInfo, ce.ifFalseNullStatus, expectedNonNullness, originalExpected);
 		return;
 	}
 	int nullStatus = argument.nullStatus(flowInfo, flowContext);
-	internalAnalyseOneArgument18(currentScope, flowContext, expectedType, argument, nullStatus,
-									expectedNonNullness, originalExpected);
+	internalAnalyseOneArgument18(currentScope, flowContext, expectedType, argument, flowInfo,
+									nullStatus, expectedNonNullness, originalExpected);
 }
 void internalAnalyseOneArgument18(BlockScope currentScope, FlowContext flowContext, TypeBinding expectedType,
-		Expression argument, int nullStatus, Boolean expectedNonNullness, TypeBinding originalExpected) 
+		Expression argument, FlowInfo flowInfo, int nullStatus, Boolean expectedNonNullness, TypeBinding originalExpected) 
 {
 	// here we consume special case information generated in the ctor of ParameterizedGenericMethodBinding (see there):
 	int statusFromAnnotatedNull = expectedNonNullness == Boolean.TRUE ? nullStatus : 0;  
@@ -166,11 +188,11 @@ void internalAnalyseOneArgument18(BlockScope currentScope, FlowContext flowConte
 	if (!annotationStatus.isAnyMismatch() && statusFromAnnotatedNull != 0)
 		expectedType = originalExpected; // to avoid reports mentioning '@NonNull null'!
 	
-	if (annotationStatus.isDefiniteMismatch() || statusFromAnnotatedNull == FlowInfo.NULL) {
+	if (statusFromAnnotatedNull == FlowInfo.NULL) {
 		// immediate reporting:
 		currentScope.problemReporter().nullityMismatchingTypeAnnotation(argument, argument.resolvedType, expectedType, annotationStatus);
-	} else if (annotationStatus.isUnchecked() || (statusFromAnnotatedNull & FlowInfo.POTENTIALLY_NULL) != 0) {
-		flowContext.recordNullityMismatch(currentScope, argument, argument.resolvedType, expectedType, nullStatus);
+	} else if (annotationStatus.isAnyMismatch() || (statusFromAnnotatedNull & FlowInfo.POTENTIALLY_NULL) != 0) {
+		flowContext.recordNullityMismatch(currentScope, argument, argument.resolvedType, expectedType, flowInfo, nullStatus, annotationStatus);
 	}
 }
 
@@ -178,20 +200,20 @@ protected void checkAgainstNullTypeAnnotation(BlockScope scope, TypeBinding requ
 	if (expression instanceof ConditionalExpression && expression.isPolyExpression()) {
 		// drill into both branches using existing nullStatus per branch:
 		ConditionalExpression ce = (ConditionalExpression) expression;
-		internalCheckAgainstNullTypeAnnotation(scope, requiredType, ce.valueIfTrue, ce.ifTrueNullStatus, flowContext);
-		internalCheckAgainstNullTypeAnnotation(scope, requiredType, ce.valueIfFalse, ce.ifFalseNullStatus, flowContext);
+		internalCheckAgainstNullTypeAnnotation(scope, requiredType, ce.valueIfTrue, ce.ifTrueNullStatus, flowContext, flowInfo);
+		internalCheckAgainstNullTypeAnnotation(scope, requiredType, ce.valueIfFalse, ce.ifFalseNullStatus, flowContext, flowInfo);
 		return;
 	}
 	int nullStatus = expression.nullStatus(flowInfo, flowContext);
-	internalCheckAgainstNullTypeAnnotation(scope, requiredType, expression, nullStatus, flowContext);
+	internalCheckAgainstNullTypeAnnotation(scope, requiredType, expression, nullStatus, flowContext, flowInfo);
 }
 private void internalCheckAgainstNullTypeAnnotation(BlockScope scope, TypeBinding requiredType, Expression expression,
-		int nullStatus, FlowContext flowContext) {
+		int nullStatus, FlowContext flowContext, FlowInfo flowInfo) {
 	NullAnnotationMatching annotationStatus = NullAnnotationMatching.analyse(requiredType, expression.resolvedType, nullStatus);
 	if (annotationStatus.isDefiniteMismatch()) {
 		scope.problemReporter().nullityMismatchingTypeAnnotation(expression, expression.resolvedType, requiredType, annotationStatus);
 	} else if (annotationStatus.isUnchecked()) {
-		flowContext.recordNullityMismatch(scope, expression, expression.resolvedType, requiredType, nullStatus);
+		flowContext.recordNullityMismatch(scope, expression, expression.resolvedType, requiredType, flowInfo, nullStatus, annotationStatus);
 	}
 }
 
@@ -203,6 +225,52 @@ public void branchChainTo(BranchLabel label) {
 	// do nothing by default
 }
 
+// Inspect AST nodes looking for a break statement, descending into nested control structures only when necessary (looking for a break with a specific label.)
+public boolean breaksOut(final char[] label) {
+	return new ASTVisitor() {
+		
+		boolean breaksOut;
+		public boolean visit(TypeDeclaration type, BlockScope skope) { return label != null; }
+		public boolean visit(TypeDeclaration type, ClassScope skope) { return label != null; }
+		public boolean visit(LambdaExpression lambda, BlockScope skope) { return label != null;}
+		public boolean visit(WhileStatement whileStatement, BlockScope skope) { return label != null; }
+		public boolean visit(DoStatement doStatement, BlockScope skope) { return label != null; }
+		public boolean visit(ForeachStatement foreachStatement, BlockScope skope) { return label != null; }
+		public boolean visit(ForStatement forStatement, BlockScope skope) { return label != null; }
+		public boolean visit(SwitchStatement switchStatement, BlockScope skope) { return label != null; }
+		
+		public boolean visit(BreakStatement breakStatement, BlockScope skope) {
+			if (label == null || CharOperation.equals(label,  breakStatement.label))
+				this.breaksOut = true;
+	    	return false;
+	    }
+		
+		public boolean breaksOut() {
+			Statement.this.traverse(this, null);
+			return this.breaksOut;
+		}
+	}.breaksOut();
+}
+
+/* Inspect AST nodes looking for a continue statement with a label, descending into nested control structures.
+   The label is presumed to be NOT attached to this. This condition is certainly true for lambda shape analysis
+   where this analysis triggers only from do {} while (false); situations. See LabeledStatement.continuesAtOuterLabel
+*/
+public boolean continuesAtOuterLabel() {
+	return new ASTVisitor() {
+		boolean continuesToLabel;
+		public boolean visit(ContinueStatement continueStatement, BlockScope skope) {
+			if (continueStatement.label != null)
+				this.continuesToLabel = true;
+	    	return false;
+	    }
+		public boolean continuesAtOuterLabel() {
+			Statement.this.traverse(this, null);
+			return this.continuesToLabel;
+		}
+	}.continuesAtOuterLabel();
+}
+
 // Report an error if necessary (if even more unreachable than previously reported
 // complaintLevel = 0 if was reachable up until now, 1 if fake reachable (deadcode), 2 if fatal unreachable (error)
 public int complainIfUnreachable(FlowInfo flowInfo, BlockScope scope, int previousComplaintLevel, boolean endOfBlock) {
@@ -363,16 +431,4 @@ protected MethodBinding findConstructorBinding(BlockScope scope, Invocation site
 	resolvePolyExpressionArguments(site, ctorBinding, argumentTypes, scope);
 	return ctorBinding;
 }
-/**
- * If an exception-throwing statement is resolved within the scope of a lambda, record the exception type(s).
- * It is likely wrong to do this during resolve, should probably use precise flow information.
- */
-protected void recordExceptionsForEnclosingLambda(BlockScope scope, TypeBinding... thrownExceptions) {
-	MethodScope methodScope = scope.methodScope();
-	if (methodScope != null && methodScope.referenceContext instanceof LambdaExpression) {
-		LambdaExpression lambda = (LambdaExpression) methodScope.referenceContext;
-		for (int i = 0; i < thrownExceptions.length; i++)
-			lambda.throwsException(thrownExceptions[i]);
-	}
-}
 }
