@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2000, 2014 IBM Corporation and others.
+ * Copyright (c) 2000, 2015 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -21,6 +21,9 @@
  *								Bug 439516 - [1.8][null] NonNullByDefault wrongly applied to implicit type bound of binary type
  *								Bug 438458 - [1.8][null] clean up handling of null type annotations wrt type variables
  *								Bug 435570 - [1.8][null] @NonNullByDefault illegally tries to affect "throws E"
+ *								Bug 435805 - [1.8][compiler][null] Java 8 compiler does not recognize declaration style null annotations
+ *								Bug 437072 - [compiler][null] Null analysis emits possibly incorrect warning for new int[][] despite @NonNullByDefault
+ *								Bug 466713 - Null Annotations: NullPointerException using <int @Nullable []> as Type Param 
  *        Andy Clement (GoPivotal, Inc) aclement@gopivotal.com - Contributions for
  *                          Bug 383624 - [1.8][compiler] Revive code generation support for type annotations (from Olivier's work)
  *                          Bug 409236 - [1.8][compiler] Type annotations on intersection cast types dropped by code generator
@@ -33,7 +36,6 @@ import java.util.List;
 
 import org.aspectj.org.eclipse.jdt.internal.compiler.ASTVisitor;
 import org.aspectj.org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching.CheckMode;
-import org.aspectj.org.eclipse.jdt.internal.compiler.classfmt.ClassFileConstants;
 import org.aspectj.org.eclipse.jdt.internal.compiler.codegen.AnnotationContext;
 import org.aspectj.org.eclipse.jdt.internal.compiler.codegen.AnnotationTargetTypeConstants;
 import org.aspectj.org.eclipse.jdt.internal.compiler.flow.FlowContext;
@@ -59,6 +61,24 @@ import org.aspectj.org.eclipse.jdt.internal.compiler.problem.ProblemSeverities;
 @SuppressWarnings({"rawtypes", "unchecked"})
 public abstract class TypeReference extends Expression {
 	public static final TypeReference[] NO_TYPE_ARGUMENTS = new TypeReference[0];
+
+	/**
+	 * Simplified specification of where in a (possibly complex) type reference
+	 * we are looking for type annotations.
+	 * @see TypeReference#hasNullTypeAnnotation(AnnotationPosition)
+	 */
+	public static enum AnnotationPosition {
+		/**
+		 * For arrays: the outermost dimension, for parameterized types the type, for nested types the innermost type.
+		 * This is the level that a declaration annotation would apply to.
+		 */
+		MAIN_TYPE,
+		/** For arrays: the leaf component type, else like MAIN_TYPE. */
+		LEAF_TYPE,
+		/** Any position admitting type annotations. */
+		ANY
+	}
+	
 static class AnnotationCollector extends ASTVisitor {
 	List annotationContexts;
 	Expression typeReference;
@@ -659,9 +679,7 @@ public int getAnnotatableLevels() {
 }
 /** Check all typeArguments against null constraints on their corresponding type variables. */
 protected void checkNullConstraints(Scope scope, TypeReference[] typeArguments) {
-	CompilerOptions compilerOptions = scope.compilerOptions();
-	if (compilerOptions.isAnnotationBasedNullAnalysisEnabled
-			&& compilerOptions.sourceLevel >= ClassFileConstants.JDK1_8
+	if (scope.environment().usesNullTypeAnnotations()
 			&& typeArguments != null)
 	{
 		TypeVariableBinding[] typeVariables = this.resolvedType.original().typeVariables();
@@ -681,6 +699,9 @@ protected void checkNullConstraints(Scope scope, TypeBinding[] variables, int ra
 				scope.problemReporter().nullityMismatchTypeArgument(variable, this.resolvedType, this);
     	}
 	}
+	if (this.resolvedType.leafComponentType().isBaseType() && hasNullTypeAnnotation(AnnotationPosition.LEAF_TYPE)) {
+		scope.problemReporter().illegalAnnotationForBaseType(this, this.annotations[0], this.resolvedType.tagBits & TagBits.AnnotationNullMASK);
+	}
 }
 /** Retrieve the null annotation that has been translated to the given nullTagBits. */
 public Annotation findAnnotation(long nullTagBits) {
@@ -698,11 +719,17 @@ public Annotation findAnnotation(long nullTagBits) {
 	}
 	return null;
 }
-public boolean hasNullTypeAnnotation() {
+public boolean hasNullTypeAnnotation(AnnotationPosition position) {
 	if (this.annotations != null) {
+		if (position == AnnotationPosition.MAIN_TYPE) {
 		Annotation[] innerAnnotations = this.annotations[this.annotations.length-1];
-		if (containsNullAnnotation(innerAnnotations))
+			return containsNullAnnotation(innerAnnotations);
+		} else {
+			for (Annotation[] someAnnotations: this.annotations) {
+				if (containsNullAnnotation(someAnnotations))
 			return true;
+	}
+		}
 	}
 	return false;
 }
@@ -721,4 +748,8 @@ public static boolean containsNullAnnotation(Annotation[] annotations) {
 public TypeReference[] getTypeReferences() {
 	return new TypeReference [] { this };
 }
+
+public boolean isBaseTypeReference() {
+	return false;
+}
 }
