@@ -108,6 +108,12 @@ public class NonNullDefaultAwareTypeAnnotationWalker extends TypeAnnotationWalke
 		}
 	}
 	
+	@Override
+	public ITypeAnnotationWalker toSupertype(short index, char[] superTypeSignature) {
+		if (this.isEmpty) return restrict(this.matches, this.pathPtr);
+		return super.toSupertype(index, superTypeSignature);
+	}
+
 	@Override
 	public ITypeAnnotationWalker toMethodParameter(short index) {
 		// don't set nextIsDefaultLocation, because signature-level nullness is handled by ImplicitNullAnnotationVerifier
@@ -130,6 +136,14 @@ public class NonNullDefaultAwareTypeAnnotationWalker extends TypeAnnotationWalke
 		return super.toTypeBound(boundIndex);
 	}
 
+	@Override
+	public ITypeAnnotationWalker toWildcardBound() {
+		this.nextIsDefaultLocation = (this.defaultNullness & Binding.DefaultLocationTypeBound) != 0;
+		this.nextIsTypeBound = true;
+		if (this.isEmpty) return restrict(this.matches, this.pathPtr);
+		return super.toWildcardBound();
+	}
+
 	@Override
 	public ITypeAnnotationWalker toTypeParameterBounds(boolean isClassTypeParameter, int parameterRank) {
 		this.nextIsDefaultLocation = (this.defaultNullness & Binding.DefaultLocationTypeBound) != 0;
@@ -154,6 +168,12 @@ public class NonNullDefaultAwareTypeAnnotationWalker extends TypeAnnotationWalke
 		return super.toTypeParameter(isClassTypeParameter, rank);
 	}
 
+	@Override
+	protected ITypeAnnotationWalker toNextDetail(int detailKind) {
+		if (this.isEmpty) return restrict(this.matches, this.pathPtr);
+		return super.toNextDetail(detailKind);
+	}
+
 	@Override
 	public IBinaryAnnotation[] getAnnotationsAtCursor(int currentTypeId) {
 		IBinaryAnnotation[] normalAnnotations = this.isEmpty ? null : super.getAnnotationsAtCursor(currentTypeId);
