@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2005, 2014 IBM Corporation and others.
+ * Copyright (c) 2005, 2015 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -9,6 +9,7 @@
  *     IBM Corporation - initial API and implementation
  *     Stephan Herrmann - Contribution for
  *     							Bug 425183 - [1.8][inference] make CaptureBinding18 safe
+ *								Bug 466308 - [hovering] Javadoc header for parameter is wrong with annotation-based null analysis
  *******************************************************************************/
 package org.aspectj.org.eclipse.jdt.internal.core.util;
 
@@ -18,6 +19,7 @@ import org.aspectj.org.eclipse.jdt.core.Signature;
 import org.aspectj.org.eclipse.jdt.core.compiler.CharOperation;
 import org.aspectj.org.eclipse.jdt.internal.compiler.ASTVisitor;
 import org.aspectj.org.eclipse.jdt.internal.compiler.Compiler;
+import org.aspectj.org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
 import org.aspectj.org.eclipse.jdt.internal.compiler.ast.ArrayReference;
 import org.aspectj.org.eclipse.jdt.internal.compiler.ast.Assignment;
 import org.aspectj.org.eclipse.jdt.internal.compiler.ast.CastExpression;
@@ -55,6 +57,53 @@ import org.aspectj.org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
 
 @SuppressWarnings({"rawtypes", "unchecked"})
 public class BindingKeyResolver extends BindingKeyParser {
+
+	/** Synthetic bindings for local variables (method arguments) restored from a binding key. */
+	private final class SyntheticLocalVariableBinding extends LocalVariableBinding {
+
+		private final MethodBinding enclosingMethod;
+		private int paramPosition;
+		private char[] key;
+
+		SyntheticLocalVariableBinding(char[] name, TypeBinding type, MethodBinding enclosingMethod, int paramPosition) {
+			super(name, type, 0, true);
+			this.enclosingMethod = enclosingMethod;
+			this.paramPosition = paramPosition;
+		}
+
+		@Override
+		public char[] computeUniqueKey() {
+			if (this.key == null) {
+				// have no scope to find the enclosing method, so use the captured method:
+				StringBuilder buf = new StringBuilder().append(this.enclosingMethod.computeUniqueKey());
+				buf.append('#');
+				buf.append(this.name);
+				buf.append("#0#"); //$NON-NLS-1$
+				buf.append(this.paramPosition);
+				int length = buf.length();
+				this.key = new char[length];
+				buf.getChars(0, length, this.key, 0);
+			}
+			return this.key;
+		}
+		
+		@Override
+		public MethodBinding getEnclosingMethod() {
+			return this.enclosingMethod;
+		}
+
+		@Override
+		public int hashCode() {
+			return CharOperation.hashCode(computeUniqueKey());
+		}
+
+		public boolean equals(Object obj) {
+			if (!(obj instanceof SyntheticLocalVariableBinding))
+				return false;
+			return CharOperation.equals(computeUniqueKey(), ((SyntheticLocalVariableBinding) obj).computeUniqueKey());
+		}
+	}
+
 	Compiler compiler;
 	Binding compilerBinding;
 
@@ -184,7 +233,7 @@ public class BindingKeyResolver extends BindingKeyParser {
 					case Binding.TYPE_PARAMETER:
 						if (binding.isCapture()) {
 							CaptureBinding captureBinding = (CaptureBinding) binding;
-							if (captureBinding.position == position && captureBinding.wildcard == wildcardBinding) {
+							if (captureBinding.end == position && captureBinding.wildcard == wildcardBinding) {
 								if (captureBinding instanceof CaptureBinding18) {
 									if (((CaptureBinding18)captureBinding).captureID != capture18id)
 										return false;
@@ -286,19 +335,43 @@ public class BindingKeyResolver extends BindingKeyParser {
  			}
 	}
 
-	public void consumeLocalVar(char[] varName, int occurrenceCount) {
+	public void consumeLocalVar(char[] varName, int occurrenceCount, int argumentPosition) {
 		if (this.scope == null) {
 			if (this.methodBinding == null)
 				return;
-			this.scope = this.methodBinding.sourceMethod().scope;
+			AbstractMethodDeclaration sourceMethod = this.methodBinding.sourceMethod();
+			if (sourceMethod != null) {
+				this.scope = sourceMethod.scope;
+			} else {
+				char[][] parameterNames = this.methodBinding.parameterNames;
+				int paramPosition = -1;
+				if (parameterNames.length == 0) {
+					paramPosition = argumentPosition;
+				} else {
+					for (int i = 0; i < parameterNames.length; i++) {
+						if (CharOperation.equals(parameterNames[i], varName)) {
+							paramPosition = i;
+							break;
+						}
+					}
+				}
+				if (paramPosition != -1) {
+					// we don't have a compiler binding for this argument, but we can craft one:
+					this.compilerBinding = new SyntheticLocalVariableBinding(varName, this.methodBinding.parameters[paramPosition], this.methodBinding, paramPosition);
+					this.methodBinding = null;
+					return;
+				}
+			}
 		}
-	 	for (int i = 0; i < this.scope.localIndex; i++) {
-			LocalVariableBinding local = this.scope.locals[i];
-			if (CharOperation.equals(local.name, varName)
-					&& occurrenceCount-- == 0) {
-				this.methodBinding = null;
-				this.compilerBinding = local;
-				return;
+		if (this.scope != null) {
+		 	for (int i = 0; i < this.scope.localIndex; i++) {
+				LocalVariableBinding local = this.scope.locals[i];
+				if (CharOperation.equals(local.name, varName)
+						&& occurrenceCount-- == 0) {
+					this.methodBinding = null;
+					this.compilerBinding = local;
+					return;
+				}
 			}
 		}
 	}
